import base64
import requests
from sermonaudio import get_base_url, get_request_headers
from posixpath import join as posixjoin

URL_PATH_BROADCASTER = 'broadcaster'


def create_sermon(full_title: str,
                  speaker_name: str,
                  preach_date: str,
                  accept_additional_charges: bool,
                  display_title: str = None,
                  subtitle: str = '',
                  bible_text: str = '',
                  more_info_text: str = '',
                  event_type: str = 'Sunday Service',
                  language_code: str = 'en',
                  keywords: str = ''):
    """Creates a sermon record."""
    data = {
        'acceptAdditionalCharges': accept_additional_charges,
        'fullTitle': full_title,
        'displayTitle': display_title,
        'subtitle': subtitle,
        'speakerName': speaker_name,
        'preachDate': preach_date,
        'bibleText': bible_text,
        'eventType': event_type,
        'languageCode': language_code,
        'moreInfoText': more_info_text,
        'keywords': ' '.join(keywords)
        }

    broadcaster_create_endpoint = 'create_sermon'

    url = posixjoin(get_base_url(), URL_PATH_BROADCASTER, broadcaster_create_endpoint)

    r = requests.post(url, data=data, headers=get_request_headers())

    response_json = r.json()

    if 'errors' in response_json:
        return False, str(response_json), None
    else:
        message = response_json['message'] if 'message' in response_json.keys() else None
        return response_json['success'], message, response_json['sermonID']


def upload_audio(sermon_id: str,
                 filename,
                 file_content):  # pragma: no cover
    """Attaches audio to a sermon.

    Pass in a bytes-like object for file content. The result is automatically encoded."""

    data = {'sermonID': sermon_id,
            'filename': filename,
            'fileData': base64.b64encode(file_content).decode('utf-8')
            }

    broadcaster_upload_endpoint = 'upload_audio'
    url = posixjoin(get_base_url(), URL_PATH_BROADCASTER, broadcaster_upload_endpoint)

    r = requests.post(url,
                      json=data,
                      headers=get_request_headers())
    
    return r.status_code == 200



import requests

from posixpath import join as posixjoin

import dateutil.parser

from sermonaudio import get_base_url, get_request_headers
from sermonaudio.node import types

URL_PATH = 'node'

def _get_sermons_node(node_name: str, params: dict, preferred_language_override=None):
    """Calls an endpoint named node_name with params, and expects to get a list of sermons

    :param node_name: The name of the node to call
    :param params: A dictionary of query string parameters
    :param preferred_language_override: A preferred langauge code, if you don't want to use your system locale
    :rtype: [sermonaudio.types.Sermon]
    """
    url = posixjoin(get_base_url(), URL_PATH, node_name)
    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))

    node = types.Node(response.json())
    assert node.node_type == node_name

    sermons = [types.Sermon(result) for result in node.results]
    for sermon in sermons:
        assert isinstance(sermon, types.Sermon)

    return sermons


def get_sermon_info(sermon_id: str, preferred_language_override=None):
    """Calls the sermon_info node endpoint.

    :param sermon_id: The ID of the sermon you want to fetch
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: sermonaudio.types.Sermon or None
    """
    node_name = 'sermon_info'
    params = {
        'sermonID': sermon_id
    }
    sermons = _get_sermons_node(node_name, params, preferred_language_override)

    if len(sermons) > 0:
        return sermons[0]
    else:
        return None


def get_newest_audio_sermons(page: int = 1,
                             page_size: int = None,
                             preferred_language_override=None):
    """Calls the newest_audio_sermons node endpoint.

    Returned sermons are sorted by media upload date descending.

    :param page: The page number to load (defaults to 1).
    :param page_size: The number of items per page (currently defaults to 50 if omitted).
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: [sermonaudio.types.Sermon]
    """
    node_name = 'newest_audio_sermons'
    params = {
        'page': page,
        'page_size': page_size
    }

    return _get_sermons_node(node_name, params, preferred_language_override)


def get_newest_video_sermons(page: int = 1,
                             page_size: int = None,
                             preferred_language_override=None):
    """Calls the newest_video_sermons node endpoint.

    Returned sermons are sorted by media upload date descending.

    :param page: The page number to load (defaults to 1)
    :param page_size: The number of items per page (currently defaults to 50 if omitted).
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: [sermonaudio.types.Sermon]
    """
    node_name = 'newest_video_sermons'
    params = {
        'page': page,
        'page_size': page_size
    }

    return _get_sermons_node(node_name, params, preferred_language_override)


def get_sermons_by_source(source_id: str,
                          only_audio_ready: bool = False,
                          event_type: str = None,
                          series: str = None,
                          speaker_name: str = None,
                          sermon_type: str = None,
                          page: int = 1,
                          page_size: int = None,
                          preferred_language_override=None):
    """Calls the sermons_by_source node endpoint.

    Returned sermons are sorted by date descending.

    :param source_id: The ID of the source you want a listing from
    :param only_audio_ready: If set, this flag causes the endpoint to
    return only sermons with audio available (default is to return all sermons)
    :param event_type: Filter results by sermon event type (used only if a value is supplied)
    :param series: Filter results by series (used only if a value is supplied)
    :param speaker_name: Filter results by speaker name (used only if a value is supplied)
    :param sermon_type: Filter results by type of sermon (e.g., audio or video; used only if a value is supplied)
    :param page: The page number to load (defaults to 1)
    :param page_size: The number of items per page (currently defaults to 50 if omitted).
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: [sermonaudio.types.Sermon]
    """
    node_name = 'sermons_by_source'
    params = {
        'sourceID': source_id,
        'page': page,
        'page_size': page_size,
        'only_audio_ready': only_audio_ready,
        'sermon_type': sermon_type,
        'event_type': event_type,
        'series': series,
        'speaker_name': speaker_name,
    }

    return _get_sermons_node(node_name, params, preferred_language_override)


def get_sermons_by_speaker(speaker_name: str,
                           event_type: str = None,
                           series: str = None,
                           page: int = 1,
                           page_size: int = None,
                           preferred_language_override=None):
    """Calls the sermons_by_speaker node endpoint.

    Returned sermons are sorted by date descending.

    :param speaker_name: The name of the speaker
    :param event_type: Filter results by sermon event type (used only if a value is supplied)
    :param series: Filter results by series (used only if a value is supplied)
    :param page: The page number of results to load (defaults to 1)
    :param page_size: The number of items per page (currently defaults to 50 if omitted).
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: list of sermonaudio.types.Sermon
    """
    node_name = 'sermons_by_speaker'
    url = posixjoin(get_base_url(), URL_PATH, node_name)
    params = {
        'page': page,
        'page_size': page_size,
        'speakerName': speaker_name,
        'event_type': event_type,
        'series': series
    }

    return _get_sermons_node(node_name, params, preferred_language_override)


def get_sermons_by_bibref(bibbook: str,
                          bibchapter: int = None,
                          bibchapterend: int = None,
                          bibverse: int = None,
                          bibverseend: int = None,
                          page: int = 1,
                          page_size: int = None,
                          preferred_language_override=None):

    """Lists sermons preached on the specific text supplied. Text can
    either be specific, or in a range from bibchapter:bibverse to
    bibchapterend:bibverseend (with searches not spanning across
    books).

    Note that bibref does not use preferred_language_override, because
    the bibbook parameter is an OSIS paratext abbreviation and never
    localized.

    :param bibbook: The book of the Bible to search for (in three-letter OSIS paratext abbreviation format)
    :param bibchapter: Optional, limits results to this chapter of the book
    :param bibchapterend: Optional, indicates a range from bibchapterto bibchapterend
    :param bibverse: Optional, limits the results
    :param bibverseend: Optional, indicates a range from bibverse to bibverseend
    :param page: The page number of results to load (defaults to 1)
    :param page_size: The number of items per page (currently defaults to 50 if omitted).
    :rtype: list of sermonaudio.types.Sermon
    """

    node_name = 'sermons_by_bibref'

    # these aren't optional/have defaults...
    params = {
        'page': page,
        'page_size': page_size,
        'bibbook': bibbook
    }
    
    #... but these are/do
    if bibchapter:
        params['bibchapter'] = bibchapter
    if bibchapterend:
        params['bibchapterend'] = bibchapterend
    if bibverse:
        params['bibverse'] = bibverse
    if bibverseend:
        params['bibverseend'] = bibverseend

    return _get_sermons_node(node_name, params, None)


def get_sermons_by_language(language_code: str,
                            page: int = 1,
                            page_size: int = None,
                            preferred_language_override=None):
    """Calls the sermons_by_language node endpoint.

    Returned sermons are sorted by date descending.

    :param language_code: An ISO 639 language code
    :param page: The page number to load (defaults to 1)
    :param page_size: The number of items per page (currently defaults to 50 if omitted).
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: [sermonaudio.types.Sermon]
    """
    node_name = 'sermons_by_language'
    params = {
        'languageCode': language_code,
        'page': page,
        'page_size': page_size
    }

    return _get_sermons_node(node_name, params, preferred_language_override)


def get_source_info(source_id: str, preferred_language_override=None):
    """Calls the source_info node endpoint.

    :param source_id: The ID of the source you want info on.
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: sermonaudio.types.Source
    """
    node_name = 'source_info'
    url = posixjoin(get_base_url(), URL_PATH, node_name)
    params = {
        'sourceID': source_id
    }

    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))

    node = types.Node(response.json())
    assert node.node_type == node_name

    if len(node.results) > 0:
        source = types.Source(node.results[0])
        assert isinstance(source, types.Source)

        return source
    else:
        return None


def get_webcasts_in_progress(source_id: str = None,
                             preferred_language_override=None):
    """Calls the webcasts_in_progress node endpoint.

    :param source_id: You can omit this since it defaults to API key.
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: [sermonaudio.types.WebcastInfo]
    """
    node_name = 'webcasts_in_progress'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {}
    if source_id:
        params['sourceID'] = source_id

    response = requests.get(url,
                            params=params,
                            headers=get_request_headers(preferred_language_override))

    node = types.Node(response.json())
    assert node.node_type == node_name

    webcasts = [types.WebcastInfo(result) for result in node.results]

    for webcast in webcasts:
        assert isinstance(webcast, types.WebcastInfo)

    return webcasts


def get_sources_near_location(latitude: float, longitude: float, meters: int, preferred_language_override=None):
    """Calls the source_info node endpoint.

    :param latitude: The latitude of the search origin
    :param longitude: The longitude of the search origin
    :param meters: The distance from the origin to search
    :param preferred_language_override: An optional override to the Accept-Language header for a single request
    :rtype: sermonaudio.types.Source
    """
    node_name = 'sources_near_location'
    url = posixjoin(get_base_url(), URL_PATH, node_name)
    params = {
        'latitude': latitude,
        'longitude': longitude,
        'meters': meters
    }

    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))

    node = types.Node(response.json())
    assert node.node_type == node_name

    sources_near_location = [types.SourceNearLocation(result) for result in node.results]

    for source_near_location in sources_near_location:
        assert isinstance(source_near_location, types.SourceNearLocation)

    return sources_near_location


def get_series_list_for_source(source_id: str=None, preferred_language_override=None):
    """Returns the series for the source.
    
    :param source_id: You can omit this since it defaults to API key.
    :rtype: [Series]
    """
    node_name = 'series_list'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {}
    if source_id:
        params['sourceID'] = source_id

    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    series = [types.SermonSeries(result) for result in node.results]
    assert node.node_type == 'series_list'
    return series


def get_sermon_event_list(preferred_language_override=None):
    """Returns the list of event types which can be used as a filter
    in the various get_source_by_* defs.

    :rtype: [String]
    """
    node_name = 'sermon_event_type_list'
    url = posixjoin(get_base_url(), URL_PATH, node_name)
    
    params = {}
    response = requests.get(url, params={}, headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'sermon_event_type_list'
    return node.results


def get_gallery_list(source_id: str = None):
    """Returns a list of gallery categories for the source.

    Note that preferred_language_override does not apply to this call.

    :param source_id: You can omit this since it defaults to API key.
    :rtype: [String]
    """
    node_name = 'gallery_list_categories'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {}
    if source_id:
        params['sourceID'] = source_id

    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'gallery_category_list'
    return node.results


def get_gallery_cover_image_for_category(cat: str,
                                         source_id: str = None,
                                         preferred_language_override=None):
    """Returns the cover image for a gallery, or None if there isn't one.

    Note that preferred_language_override does not apply to this call.

    :cat: Category for which to get cover image.
    :param source_id: You can omit this since it defaults to API key.
    :rtype: sermonaudio.types.GalleryItem or None
    """
    node_name = 'gallery_cover_image_for_category'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {'category': category}
    if source_id:
        params['sourceID'] = source_id

    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'gallery_cover_image'
    return types.GalleryItem(node.results) if node.results is not None else None


def get_gallery_items_for_category(cat: str,
                                   includecover: bool = True,
                                   source_id: str = None,
                                   preferred_language_override=None):
    """Returns a list of images for the gallery category, optionally
    omitting the cover image.

    Note that preferred_language_override does not apply to this call.

    :param cat: Gallery category to list.
    :param includecover: Flag indicating whether to include cover image or not.
    :param source_id: You can omit this since it defaults to API key.
    :rtype: [sermonaudio.types.GalleryItem]
    """
    node_name = 'gallery_items_for_category'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {
        'category': category,
        'include_cover_image': includecover
    }
    if source_id:
        params['sourceID'] = source_id

    response = requests.get(url, params=params,
                            headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'gallery_items_list'
    return [types.GalleryItem(item) for item in node.results]


def get_event_list(upcoming: bool = True,
                   source_id: str = None,
                   preferred_language_override=None):
    """Returns a list of events for the source. Typically, you will
    want only upcoming events, not historical ones which are already
    over, but a flag is procided to get all events.

    Note that preferred_language_override does not apply to this call.

    :param upcoming: Normally defaults to true, but can be set to false to show historical events.
    :param source_id: You can omit this since it defaults to API key.
    :rtype: [String]
    """
    node_name = 'event_list'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {'upcoming': upcoming}
    if source_id:
        params['sourceID'] = source_id

    response = requests.get(url, params=params, headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'event_list'
    return node.results


def get_event_item(event_id: str,
                   source_id: str = None,
                   preferred_language_override=None):
    """Returns an event item for a given event ID

    Note that preferred_language_override does not apply to this call.

    :param event_id: The event ID to look up.
    :param source_id: You can omit this since it defaults to API key.
    :rtype: types.EventItem
    """
    node_name = 'event_info'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {'eventID': event_id}
    if source_id:
        myparams['sourceID'] = source_id

    response = requests.get(url, params=params,
                            headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'event_info'
    return types.EventItem(node.results) if node.results is not None else None


def get_spurgeon_devotional(type: str,
                            month: int,
                            day_of_month: int,
                            preferred_language_override=None):
    """Returns the Spurgeon devotional for a given month and day
    
    :param type: The type of devotional (one of the values in types.SPURGEON_DEVOTIONAL_TYPES)
    :param month: The month to return
    :param day_of_month: The day of month to return 
    :rtype: types.SpurgeonDevotional
    """
    assert type in types.SPURGEON_DEVOTIONAL_TYPES
    
    node_name = 'spurgeon_devotional'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {
        'type': type,
        'month': month,
        'day_of_month': day_of_month
    }

    response = requests.get(url, params=params,
                            headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    assert node.node_type == 'spurgeon_devotional'
    return types.SpurgeonDevotional(node.results) if node.results is not None else None


def _get_event_source(event_id: str,
                      preferred_language_override=None):
    node_name = 'event_source'
    url = posixjoin(get_base_url(), URL_PATH, node_name)

    params = {'eventID': event_id}

    response = requests.get(url, params=params,
                            headers=get_request_headers(preferred_language_override))
    node = types.Node(response.json())
    return node.results['source']



from .requests import get_sermon_info, get_newest_audio_sermons,\
     get_newest_video_sermons, get_sermons_by_source, \
     get_sermons_by_speaker, get_sermons_by_language, \
     get_source_info, get_webcasts_in_progress, \
     get_sources_near_location, \
     get_sermons_by_bibref, \
     get_series_list_for_source, \
     get_sermon_event_list, \
     get_spurgeon_devotional, \
     get_gallery_list, get_gallery_cover_image_for_category, get_gallery_items_for_category, \
     get_event_list, get_event_item



"""Node API data structures"""
import datetime
import dateutil.parser
import pytz


SPURGEON_DEVOTIONAL_AM = 'AM' # Morning Devotional
SPURGEON_DEVOTIONAL_PM = 'PM' # Evening Devotional
SPURGEON_DEVOTIONAL_CHECKBOOK = 'CHECKBOOK' # Faith's Checkbook (note American spelling)
SPURGEON_DEVOTIONAL_TYPES = {SPURGEON_DEVOTIONAL_AM, SPURGEON_DEVOTIONAL_PM, SPURGEON_DEVOTIONAL_CHECKBOOK}


def isinstance_or_none(obj, type):
    return obj is None or isinstance(obj, type)


class Node(object):
    """The base node object, which encapsulates all node API responses"""
    def __init__(self, obj: dict):
        self.node_type = obj['nodeType']
        assert isinstance(self.node_type, str)

        self.node_display_name = obj['nodeDisplayName']
        assert isinstance(self.node_display_name, str)

        self.results = obj['results']
        assert (isinstance(self.results, list) or
                isinstance(self.results, dict) or
                self.results is None)

        self.total_count = obj['totalCount']
        assert isinstance_or_none(self.total_count, int)


class Source(object):
    """Broadcaster API data model"""
    def __init__(self, obj: dict):
        """Constructs a broadcaster from a dictionary returned from the SermonAudio API"""
        self.source_id = obj['sourceID']
        assert isinstance(self.source_id, str)

        self.service_times_are_preformatted = obj['serviceTimesArePreformatted']
        assert isinstance_or_none(self.service_times_are_preformatted, bool)

        self.service_times = obj['serviceTimes']
        assert isinstance_or_none(self.service_times, str)

        self.denomination = obj['denomination']
        assert isinstance_or_none(self.denomination, str)

        self.address = obj['address']
        assert isinstance_or_none(self.address, str)

        self.display_name = obj['displayName']
        assert isinstance(self.display_name, str)

        self.location = obj['location']
        assert isinstance(self.location, str)

        self.latitude = obj['latitude']
        assert isinstance_or_none(self.latitude, float)

        self.longitude = obj['longitude']
        assert isinstance_or_none(self.longitude, float)

        self.image_url = obj['imageURL']
        assert isinstance(self.image_url, str)

        self.album_art_url_format = obj['albumArtURL']
        assert isinstance(self.album_art_url_format, str)

        self.minister = obj['minister']
        assert isinstance_or_none(self.minister, str)

        self.phone = obj['phone']
        assert isinstance_or_none(self.phone, str)

        self.home_page_url = obj['homePageURL']
        assert isinstance_or_none(self.home_page_url, str)

        self.bible_version = obj['bibleVersion']
        assert isinstance_or_none(self.bible_version, str)

        self.facebook_username = obj['facebookUsername']
        assert isinstance_or_none(self.facebook_username, str)

        self.twitter_username = obj['twitterUsername']
        assert isinstance_or_none(self.twitter_username, str)

        self.about_us = obj['aboutUs']
        assert isinstance_or_none(self.about_us, str)

        self.webcast_in_progress = obj['webcastInProgress']
        assert isinstance(self.webcast_in_progress, bool)

        self.vacant_pulpit = obj['vacantPulpit']
        assert isinstance(self.vacant_pulpit, bool)

        self.disabled = obj['disabled']
        assert isinstance(self.disabled, bool)

    def get_album_art_url(self, size: int):
        """Returns a URL for the square album art with a with and height equal to the provided size argument"""
        return self.album_art_url_format.replace('{size}', str(size))


class Speaker(object):
    """Speaker data model"""
    def __init__(self, obj: dict):
        self.display_name = obj['displayName']
        assert isinstance(self.display_name, str)

        self.sort_name = obj.get('sortName')
        assert isinstance(self.sort_name, str)

        self.bio = obj['bio']
        assert isinstance_or_none(self.bio, str)

        self.portrait_url = obj['portraitURL']
        assert isinstance(self.portrait_url, str)

        self.rounded_thumbnail_image_url = obj['roundedThumbnailImageURL']
        assert isinstance(self.rounded_thumbnail_image_url, str)

        self.album_art_url_format = obj['albumArtURL']
        assert isinstance(self.album_art_url_format, str)

    def get_album_art_url(self, size: int):
        """Returns a URL for the square album art with a with and height equal to the provided size argument"""
        return self.album_art_url_format.replace('{size}', str(size))


class Sermon(object):
    """Sermon data model"""
    def __init__(self, obj: dict):
        """Constructs a sermon from a dictionary returned from the SermonAudio API"""
        self.sermon_id = obj['sermonID']
        assert isinstance(self.sermon_id, str)

        self.source = Source(obj['source'])
        self.speaker = Speaker(obj['speaker'])

        self.audio_duration_in_seconds = obj['audioDurationInSeconds']
        assert isinstance_or_none(self.audio_duration_in_seconds, int)

        self.video_duration_in_seconds = obj['videoDurationInSeconds']
        assert isinstance_or_none(self.video_duration_in_seconds, int)

        self.full_title = obj['fullTitle']
        assert isinstance(self.full_title, str)

        self.display_title = obj['displayTitle']
        assert isinstance(self.display_title, str)

        self.subtitle = obj['subtitle']
        assert isinstance_or_none(self.subtitle, str)

        self.preach_date = datetime.datetime.strptime(obj['preachDate'], '%Y-%m-%d').date()
        assert isinstance(self.preach_date, datetime.date)

        self.language_code = obj['languageCode']
        assert isinstance(self.language_code, str)

        self.bible_text = obj['bibleText']
        assert isinstance_or_none(self.bible_text, str)

        self.more_info_text = obj['moreInfoText']
        assert isinstance_or_none(self.more_info_text, str)

        self.event_type = obj['eventType']
        assert isinstance(self.event_type, str)

        self.download_count = obj['downloadCount']
        assert isinstance(self.download_count, int)

        self.audio_bitrate = obj['audioBitrate']
        assert isinstance_or_none(self.audio_bitrate, int)

        self.video_bitrate = obj['videoBitrate']
        assert isinstance_or_none(self.video_bitrate, int)

        self.audio_file_url = obj['audioFileURL']
        assert isinstance_or_none(self.audio_file_url, str)

        self.audio_file_lo_url = obj['audioFileLOURL']
        assert isinstance_or_none(self.audio_file_lo_url, str)

        self.video_file_url = obj['videoFileURL']
        assert isinstance_or_none(self.video_file_url, str)

        self.video_file_lo_url = obj['videoFileLOURL']
        assert isinstance_or_none(self.video_file_lo_url, str)

        self.video_hls_url = obj['videoHLSURL']
        assert isinstance_or_none(self.video_hls_url, str)

        self.video_thumbnail_image_url = obj['videoThumbnailImageURL']
        assert isinstance_or_none(self.video_thumbnail_image_url, str)

        self.text_file_url = obj['textFileURL']
        assert isinstance_or_none(self.text_file_url, str)

    def __str__(self):
        return '{0:s} - {1:s}'.format(self.speaker.display_name, self.display_title)


class HLSStreamInfo(object):
    """An HLS video stream info object"""

    def __init__(self, bitrate: int, url: str):
        """Constructs an HLS stream info object"""
        self.bitrate = bitrate
        assert isinstance(self.bitrate, int)

        self.url = url
        assert isinstance(self.url, str)


class WebcastInfo(object):
    """Webcast info data model"""

    def __init__(self, obj: dict):
        """Constructs webcast info from a dictionary returned from the SermonAudio API"""
        self.display_name = obj['displayName']
        assert isinstance(self.display_name, str)

        self.source_id = obj['sourceID']
        assert isinstance(self.source_id, str)

        self.source_location = obj['sourceLocation']
        assert isinstance(self.source_location, str)

        self.start_time = datetime.datetime.fromtimestamp(obj['startTime'], tz=pytz.utc)
        assert isinstance(self.start_time, datetime.datetime)

        self.preview_image_url = obj['previewImageURL']
        assert isinstance_or_none(self.preview_image_url, str)

        self.resizable_preview_image_url = obj['resizablePreviewImageURL']
        assert isinstance_or_none(self.resizable_preview_image_url, str)

        self.peak_listener_count = obj['peakListenerCount']
        assert isinstance(self.peak_listener_count, int)

        self.total_tune_in_count = obj['totalTuneInCount']
        assert isinstance(self.total_tune_in_count, int)

        self.has_video = obj['hasVideo']
        assert isinstance(self.has_video, bool)

        self.audio_url = obj['audioURL']
        assert isinstance(self.audio_url, str)

        self.hls_video_streams = []
        for stream in obj['hlsVideoStreams']:
            stream_info = HLSStreamInfo(stream['bitrate'], stream['url'])
            self.hls_video_streams.append(stream_info)


class SourceNearLocation(object):
    """Model representing a source and it's distance from a search origin"""

    def __init__(self, obj: dict):
        """Constructs source near location from a dictionary returned from the SermonAudio API"""
        self.source = Source(obj['source'])
        assert isinstance(self.source, Source)

        self.meters = obj['meters']
        assert isinstance(self.meters, int)


class GalleryItem(object):
    """One item in a gallery category, representing an image."""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.image = obj['image']
        self.category = obj['category']
        self.width = obj['width']
        self.height = obj['height']
        self.caption = obj['caption']
        self.display_caption = obj['displayCaption']
        self.url = obj['url']
        self.url_thumbnail = obj['urlThumbnail']


class EventItem(object):
    """An event item"""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.event_id = obj['eventID']
        self.title = obj['title']
        self.subtitle = obj['subtitle']
        self.category = obj['category']
        self.date_begin = dateutil.parser.parse(obj['dateBegin']).date() if obj['dateBegin'] is not None else None
        self.date_end = dateutil.parser.parse(obj['dateEnd']).date() if obj['dateEnd'] is not None else None
        self.time = obj['time']
        self.who = obj['who']
        self.cost = obj['cost']
        self.currency_type = obj['currencyType']
        self.url = obj['url']
        self.location_url = obj['locationUrl']
        self.video_url = obj['videoUrl']
        self.location_info = obj['locationInfo']
        self.location_geo_info_object = obj['locationGeoInfoObject']
        self.description = obj['description']
        self.notes = obj['notes']
        self.max_seats = obj['maxSeats']
        self.location_geo_info = obj['locationGeoInfo']
        self.accept_registration = obj['acceptRegistration']
        self.ec_flags = obj['ecFlags']

        # contained objects
        self.source = Source(obj['source'])
        self.background_image = GalleryItem(obj['backgroundImage']) if obj['backgroundImage'] is not None else None
        self.gallery = [GalleryItem(g) for g in obj['gallery'] if g is not None]
        self.speakers = [EventSpeaker(s) for s in obj['speakers']]
        self.schedule = [EventSchedule(s) for s in obj['schedule']]

        assert isinstance_or_none(self.background_image, GalleryItem)

        if self.gallery:
            for g in self.gallery:
                assert isinstance_or_none(g, GalleryItem)

        if self.speakers:
            for s in self.speakers:
                assert isinstance_or_none(s, EventSpeaker)

        if self.schedule:
            for s in self.schedule:
                assert isinstance_or_none(s, EventSchedule)


class EventSpeaker(object):
    """An event speaker"""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.sa_speaker_info = obj['saSpeakerInfo']
        self.speaker_rank = obj['speakerRank']
        self.speaker_name = obj['speakerName']


class EventNote(object):
    """An event note"""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.title = obj['title']
        self.description = obj['description']
        self.rank = obj['rank']


class EventSchedule(object):
    """An event schedule item"""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.date_begin = dateutil.parser.parse(obj['dateBegin']) if obj['dateBegin'] is not None else None
        self.date_end = dateutil.parser.parse(obj['dateEnd']) if obj['dateEnd'] is not None else None
        self.title = obj['title']
        self.subtitle = obj['subtitle']
        self.notes = obj['notes']


class SermonSeries(object):
    """A single object representing a sermon series"""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.latest = dateutil.parser.parse(obj['latest']) if obj['latest'] is not None else None
        self.earliest = dateutil.parser.parse(obj['earliest']) if obj['earliest'] is not None else None
        self.title = obj['title']
        self.count = obj['count']


class SpurgeonDevotional(object):
    """An instance of a Spurgeon devotional"""
    def __init__(self,
                 obj: dict):
        self.type = obj.get('type')
        self.month = obj['month']
        self.day = obj['day']
        self.type = obj['type']
        self.quote = obj['quote']
        self.reference = obj['reference']
        self.content = obj['content']
        self.audio = Sermon(obj['audio'])
